#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 18
void NES_mapper18::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  for(int i = 0; i < sizeof(regs)/sizeof(regs[0]); i++)
    regs[i] = 0;

  irq_enabled = 0;
  irq_latch = 0;
  irq_counter = 0;
}

void NES_mapper18::MemoryWrite(uint32 addr, uint8 data)
{
  // regs[ 0] ... 8K PROM bank at CPU $8000
  // regs[ 1] ... 8K PROM bank at CPU $A000
  // regs[ 2] ... 8K PROM bank at CPU $C000
  // regs[ 3] ... 1K VROM bank at PPU $0000
  // regs[ 4] ... 1K VROM bank at PPU $0400
  // regs[ 5] ... 1K VROM bank at PPU $0800
  // regs[ 6] ... 1K VROM bank at PPU $0C00
  // regs[ 7] ... 1K VROM bank at PPU $1000
  // regs[ 8] ... 1K VROM bank at PPU $1400
  // regs[ 9] ... 1K VROM bank at PPU $1800
  // regs[10] ... 1K VROM bank at PPU $1C00

  switch(addr)
  {
    case 0x8000:
      {
        regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
        set_CPU_bank4(regs[0]);
      }
      break;

    case 0x8001:
      {
        regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
        set_CPU_bank4(regs[0]);
      }
      break;

    case 0x8002:
      {
        regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
        set_CPU_bank5(regs[1]);
      }
      break;

    case 0x8003:
      {
        regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
        set_CPU_bank5(regs[1]);
      }
      break;

    case 0x9000:
      {
        regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
        set_CPU_bank6(regs[2]);
      }
      break;

    case 0x9001:
      {
        regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
        set_CPU_bank6(regs[2]);
      }
      break;

    case 0xA000:
      {
        regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
        set_PPU_bank0(regs[3]);
      }
      break;

    case 0xA001:
      {
        regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank0(regs[3]);
      }
      break;

    case 0xA002:
      {
        regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
        set_PPU_bank1(regs[4]);
      }
      break;

    case 0xA003:
      {
        regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank1(regs[4]);
      }
      break;

    case 0xB000:
      {
        regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
        set_PPU_bank2(regs[5]);
      }
      break;

    case 0xB001:
      {
        regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank2(regs[5]);
      }
      break;

    case 0xB002:
      {
        regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
        set_PPU_bank3(regs[6]);
      }
      break;

    case 0xB003:
      {
        regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank3(regs[6]);
      }
      break;

    case 0xC000:
      {
        regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
        set_PPU_bank4(regs[7]);
      }
      break;

    case 0xC001:
      {
        regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank4(regs[7]);
      }
      break;

    case 0xC002:
      {
        regs[8] = (regs[8] & 0xF0) | (data & 0x0F);
        set_PPU_bank5(regs[8]);
      }
      break;

    case 0xC003:
      {
        regs[8] = (regs[8] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank5(regs[8]);
      }
      break;

    case 0xD000:
      {
        regs[9] = (regs[9] & 0xF0) | (data & 0x0F);
        set_PPU_bank6(regs[9]);
      }
      break;

    case 0xD001:
      {
        regs[9] = (regs[9] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank6(regs[9]);
      }
      break;

    case 0xD002:
      {
        regs[10] = (regs[10] & 0xF0) | (data & 0x0F);
        set_PPU_bank7(regs[10]);
      }
      break;

    case 0xD003:
      {
        regs[10] = (regs[10] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank7(regs[10]);
      }
      break;

    case 0xE000:
      {
        irq_latch = (irq_latch & 0xFFF0) | (data & 0x0F);
      }
      break;

    case 0xE001:
      {
        irq_latch = (irq_latch & 0xFF0F) | ((data & 0x0F) << 4);
      }
      break;

    case 0xE002:
      {
        irq_latch = (irq_latch & 0xF0FF) | ((data & 0x0F) << 8);
      }
      break;

    case 0xE003:
      {
        irq_latch = (irq_latch & 0x0FFF) | ((data & 0x0F) << 12);
      }
      break;

    case 0xF000:
      {
        if((data & 0x01) == 0)
        {
          irq_counter = 0;
        }
        else
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xF001:
      {
        irq_enabled = data & 0x01;
      }
      break;

    case 0xF002:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else
        {
          set_mirroring(0,0,0,0);
        }
      }
      break;
  }
}

void NES_mapper18::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter <= 113)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
      irq_enabled = 0;
    }
    else
    {
      irq_counter -= 113;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 72 (Created by rinao)
void NES_mapper72::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper72::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 bank = data & 0x0f;
  if (data & 0x80)
    set_CPU_banks(bank*2, bank*2+1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  if (data & 0x40)
    set_PPU_banks(bank*8, bank*8+1,bank*8+2,bank*8+3,bank*8+4,bank*8+5,bank*8+6,bank*8+7);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 75
void NES_mapper75::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  regs[0] = 0;
  regs[1] = 1;
}

void NES_mapper75::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF000)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x9000:
      {
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        regs[0] = (regs[0] & 0x0F) | ((data & 0x02) << 3);
        set_PPU_bank0(regs[0]*4+0);
        set_PPU_bank1(regs[0]*4+1);
        set_PPU_bank2(regs[0]*4+2);
        set_PPU_bank3(regs[0]*4+3);
        regs[1] = (regs[1] & 0x0F) | ((data & 0x04) << 2);
        set_PPU_bank4(regs[1]*4+0);
        set_PPU_bank5(regs[1]*4+1);
        set_PPU_bank6(regs[1]*4+2);
        set_PPU_bank7(regs[1]*4+3);
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0xC000:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0xE000:
      {
        regs[0] = (regs[0] & 0x10) | (data & 0x0F);
        set_PPU_bank0(regs[0]*4+0);
        set_PPU_bank1(regs[0]*4+1);
        set_PPU_bank2(regs[0]*4+2);
        set_PPU_bank3(regs[0]*4+3);
      }
      break;

    case 0xF000:
      {
        regs[1] = (regs[1] & 0x10) | (data & 0x0F);
        set_PPU_bank4(regs[1]*4+0);
        set_PPU_bank5(regs[1]*4+1);
        set_PPU_bank6(regs[1]*4+2);
        set_PPU_bank7(regs[1]*4+3);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 78
void NES_mapper78::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper78::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = data & 0x0F;
  uint8 chr_bank = (data & 0xF0) >> 4;

  set_CPU_bank4(prg_bank*2);
  set_CPU_bank5(prg_bank*2+1);

  set_PPU_bank0(chr_bank*8);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);

  // added by rinao
  if((addr & 0xFE00) != 0xFE00)
  {
    if (data & 0x08)
    {
      set_mirroring(1,1,1,1);
    }
    else
    {
      set_mirroring(0,0,0,0);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 86
void NES_mapper86::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper86::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr == 0x6000)
  {
    uint8 chr_bank = data & 0x03 | (data & 0x40) >> 4;
    uint8 prg_bank = (data & 0x30) >> 4;

    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);

    set_PPU_bank0(chr_bank*8);
    set_PPU_bank1(chr_bank*8+1);
    set_PPU_bank2(chr_bank*8+2);
    set_PPU_bank3(chr_bank*8+3);
    set_PPU_bank4(chr_bank*8+4);
    set_PPU_bank5(chr_bank*8+5);
    set_PPU_bank6(chr_bank*8+6);
    set_PPU_bank7(chr_bank*8+7);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 92
void NES_mapper92::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

// Created by rinao
void NES_mapper92::MemoryWrite(uint32 addr, uint8 data)
{
  data = (uint8)addr & 0xff;
  uint8 c_bank = (data & 0x0f) << 1;
  uint8 p_bank = data & 0x0f;

  if (addr >= 0x9000){ // Moero!! ProSoccer
    if ((data & 0xf0) == 0xd0){
      set_CPU_banks(0,1,c_bank,c_bank+1);
    } else if ((data & 0xf0) == 0xe0){
      set_PPU_banks(p_bank*8,p_bank*8+1,p_bank*8+2,p_bank*8+3,p_bank*8+4,p_bank*8+5,p_bank*8+6,p_bank*8+7);
    }
  } else { // Moero!! Proyakyuu '88 ketteiban
    if ((data & 0xf0) == 0xb0){
      set_CPU_banks(0,1,c_bank,c_bank+1);
    } else if ((data & 0xf0) == 0x70){
      set_PPU_banks(p_bank*8,p_bank*8+1,p_bank*8+2,p_bank*8+3,p_bank*8+4,p_bank*8+5,p_bank*8+6,p_bank*8+7);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 101
void NES_mapper101::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper101::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  data &= 0x03;
  set_PPU_bank0(data*8);
  set_PPU_bank1(data*8+1);
  set_PPU_bank2(data*8+2);
  set_PPU_bank3(data*8+3);
  set_PPU_bank4(data*8+4);
  set_PPU_bank5(data*8+5);
  set_PPU_bank6(data*8+6);
  set_PPU_bank7(data*8+7);
}

void NES_mapper101::MemoryWrite(uint32 addr, uint8 data)
{
  data &= 0x03;
  set_PPU_bank0(data*8);
  set_PPU_bank1(data*8+1);
  set_PPU_bank2(data*8+2);
  set_PPU_bank3(data*8+3);
  set_PPU_bank4(data*8+4);
  set_PPU_bank5(data*8+5);
  set_PPU_bank6(data*8+6);
  set_PPU_bank7(data*8+7);
}
/////////////////////////////////////////////////////////////////////

#endif
