#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 19
void NES_mapper19::Reset()
{
  patch = 0;

  // for DDS Megami Tensei 2
  if(sum_16k_rom == 0x116d68)
  {
    patch = 1;
  }

  // for Mind Seeker, Kaijuu Monogatari
  if(sum_16k_rom == 0x1edbcd || sum_16k_rom == 0x1e58e7)
  {
    patch = 2;
  }

  // Init ExSound
  parent_NES->apu->SelectExSound(16);

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks >= 8)
  {
    set_PPU_bank0(num_1k_VROM_banks-8);
    set_PPU_bank1(num_1k_VROM_banks-7);
    set_PPU_bank2(num_1k_VROM_banks-6);
    set_PPU_bank3(num_1k_VROM_banks-5);
    set_PPU_bank4(num_1k_VROM_banks-4);
    set_PPU_bank5(num_1k_VROM_banks-3);
    set_PPU_bank6(num_1k_VROM_banks-2);
    set_PPU_bank7(num_1k_VROM_banks-1);
  }

  regs[0] = 0;
  regs[1] = 0;
  regs[2] = 0;
}

uint8 NES_mapper19::MemoryReadLow(uint32 addr)
{
  if(addr == 0x4800)
  {
    if(patch == 2)
    {
      uint8 retval =  parent_NES->ReadSaveRAM(regs[2] & 0x7F);
      if(regs[2] & 0x80) regs[2] = ((regs[2] & 0x7F)+1) | 0x80;
      return retval;
    }
    else
    {
      return parent_NES->apu->ExRead(addr);
    }
  }
  else if((addr & 0xF800) == 0x5000) // addr $5000-$57FF
  {
    return (uint8)(irq_counter & 0x00FF);
  }
  else if ((addr & 0xF800) == 0x5800) // addr $5800-$5FFF
  {
    return (uint8)((irq_counter & 0x7F00) >> 8);
  }
  else
  {
    return (uint8)(addr >> 8);
  }
}

void NES_mapper19::MemoryWriteLow(uint32 addr, uint8 data)
{
  switch(addr & 0xF800)
  {
    case 0x4800:
      {
        if(addr == 0x4800)
        {
          if(patch == 2)
          {
            parent_NES->WriteSaveRAM(regs[2] & 0x7F, data);
            if(regs[2] & 0x80) regs[2] = ((regs[2] & 0x7F)+1) | 0x80;
          }
          else
          {
            parent_NES->apu->ExWrite(addr, data);
          }
        }
      }
      break;

    case 0x5000: // addr $5000-$57FF
      {
        irq_counter = (irq_counter & 0xFF00) | data;
      }
      break;

    case 0x5800: // addr $5000-$57FF
      {
        irq_counter = (irq_counter & 0x00FF) | ((uint32)(data & 0x7F) << 8);
        irq_enabled = (data & 0x80) >> 7;
        if(patch)
        {
          irq_counter ++;
        }
      }
      break;
  }
}

void NES_mapper19::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF800)
  {
    case 0x8000: // addr $8000-$87FF
      {
        if(data < 0xE0 || regs[0] == 1)
        {
          set_PPU_bank0(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[0] = chr_ram + 0*0x400;
        }
      }
      break;

    case 0x8800: // addr $8800-$8FFF
      {
        if(data < 0xE0 || regs[0] == 1)
        {
          set_PPU_bank1(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[1] = chr_ram + 1*0x400;
        }
      }
      break;

    case 0x9000: // addr $9000-$97FF
      {
        if(data < 0xE0 || regs[0] == 1)
        {
          set_PPU_bank2(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[2] = chr_ram + 2*0x400;
        }
      }
      break;

    case 0x9800: // addr $9800-$9FFF
      {
        if(data < 0xE0 || regs[0] == 1)
        {
          set_PPU_bank3(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[3] = chr_ram + 3*0x400;
        }
      }
      break;

    case 0xA000: // addr $A000-$A7FF
      {
        if(data < 0xE0 || regs[1] == 1)
        {
          set_PPU_bank4(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[4] = chr_ram + 4*0x400;
        }
      }
      break;

    case 0xA800: // addr $A800-$AFFF
      {
        if(data < 0xE0 || regs[1] == 1)
        {
          set_PPU_bank5(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[5] = chr_ram + 5*0x400;
        }
      }
      break;

    case 0xB000: // addr $B000-$B7FF
      {
        if(data < 0xE0 || regs[1] == 1)
        {
          set_PPU_bank6(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[6] = chr_ram + 6*0x400;
        }
      }
      break;

    case 0xB800: // addr $B800-$BFFF
      {
        if(data < 0xE0 || regs[1] == 1)
        {
          set_PPU_bank7(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[7] = chr_ram + 7*0x400;
        }
      }
      break;

    case 0xC000: // addr $C000-$C7FF
      {
        if(data <= 0xDF)
        {
          set_PPU_bank8(data);
        }
        else
        {
          parent_NES->ppu->set_PPU_VRAM(8, data & 0x01);
        }
      }
      break;

    case 0xC800: // addr $C800-$CFFF
      {
        if(data <= 0xDF)
        {
          set_PPU_bank9(data);
        }
        else
        {
          parent_NES->ppu->set_PPU_VRAM(9, data & 0x01);
        }
      }
      break;

    case 0xD000: // addr $D000-$D7FF
      {
        if(data <= 0xDF)
        {
          set_PPU_bank10(data);
        }
        else
        {
          parent_NES->ppu->set_PPU_VRAM(10, data & 0x01);
        }
      }
      break;

    case 0xD800: // addr $D800-$DFFF
      {
        if(data <= 0xDF)
        {
          set_PPU_bank11(data);
        }
        else
        {
          parent_NES->ppu->set_PPU_VRAM(11, data & 0x01);
        }
      }
      break;

    case 0xE000: // addr $E000-$E7FF
      {
        set_CPU_bank4(data & 0x3F);
      }
      break;

    case 0xE800: // addr $E800-$EFFF
      {
        set_CPU_bank5(data & 0x3F);
        regs[0] = (data & 0x40) >> 6;
        regs[1] = (data & 0x80) >> 7;
      }
      break;

    case 0xF000: // addr $F000-$F7FF
      {
        set_CPU_bank6(data & 0x3F);
      }
      break;

    case 0xF800:
      {
        if(addr == 0xF800)
        {
          if(patch == 2)
          {
            regs[2] = data;
          }
          else
          {
            parent_NES->apu->ExWrite(addr, data);
          }
        }
      }
      break;
  }
}

void NES_mapper19::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter >= (uint32)(0x7FFF - 113))
    {
      irq_counter = 0x7FFF;
      parent_NES->cpu->DoIRQ();
    }
    else
    {
      irq_counter += 113;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 76
void NES_mapper76::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks >= 8)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper76::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr == 0x8000)
  {
    regs[0] = data;
  }
  else if(addr == 0x8001)
  {
    switch(regs[0] & 0x07)
    {
      case 0x02:
        {
          set_PPU_bank0(data*2+0);
          set_PPU_bank1(data*2+1);
        }
        break;

      case 0x03:
        {
          set_PPU_bank2(data*2+0);
          set_PPU_bank3(data*2+1);
        }
        break;

      case 0x04:
        {
          set_PPU_bank4(data*2+0);
          set_PPU_bank5(data*2+1);
        }
        break;

      case 0x05:
        {
          set_PPU_bank6(data*2+0);
          set_PPU_bank7(data*2+1);
        }
        break;

      case 0x06:
        {
          set_CPU_bank4(data);
        }
        break;

      case 0x07:
        {
          set_CPU_bank5(data);
        }
        break;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 88
void NES_mapper88::Reset()
{
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  if(num_1k_VROM_banks >= 8)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper88::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        regs[0] = data;
      }
      break;

    case 0x8001:
      {
        switch(regs[0] & 0x07)
        {
          case 0x00:
            {
              set_PPU_bank0(data & 0xFE);
              set_PPU_bank1((data & 0xFE)+1);
            }
            break;

          case 0x01:
            {
              set_PPU_bank2(data & 0xFE);
              set_PPU_bank3((data & 0xFE)+1);
            }
            break;

          case 0x02:
            {
              set_PPU_bank4(data+0x40);
            }
            break;

          case 0x03:
            {
              set_PPU_bank5(data+0x40);
            }
            break;

          case 0x04:
            {
              set_PPU_bank6(data+0x40);
            }
            break;

          case 0x05:
            {
              set_PPU_bank7(data+0x40);
            }
            break;

          case 0x06:
            {
              set_CPU_bank4(data);
            }
            break;

          case 0x07:
            {
              set_CPU_bank5(data);
            }
            break;
        }
      }
      break;

    case 0xC000:
      {
        if(data)
        {
          set_mirroring(1,1,1,1);
        }
        else
        {
          set_mirroring(0,0,0,0);
        }
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 95
void NES_mapper95::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set CPU bank pointers
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }
}

void NES_mapper95::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        regs[0] = data;
        MMC3_set_PPU_banks();
        MMC3_set_CPU_banks();
      }
      break;

    case 0x8001:
      {
        uint32 bank_num;

        if(regs[0] <= 0x05)
        {
          if(data & 0x20)
          {
            set_mirroring(1,1,1,1);
          }
          else
          {
            set_mirroring(0,0,0,0);
          }
          data &= 0x1F;
        }

        regs[1] = data;
        bank_num = regs[1];

        switch(regs[0] & 0x07)
        {
          case 0x00:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr01 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x01:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr23 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x02:
            {
              if(num_1k_VROM_banks)
              {
                chr4 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x03:
            {
              if(num_1k_VROM_banks)
              {
                chr5 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x04:
            {
              if(num_1k_VROM_banks)
              {
                chr6 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x05:
            {
              if(num_1k_VROM_banks)
              {
                chr7 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x06:
            {
              prg0 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;

          case 0x07:
            {
              prg1 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;
        }
      }
      break;
  }
}

#define MAP95_ROM(ptr)  (((ptr)-parent_NES->ROM->get_ROM_banks())  >> 13)
#define MAP95_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper95::SNSS_fixup()
{
  NES_6502::Context context;
  parent_NES->cpu->GetContext(&context);

  prg0 = MAP95_ROM(context.mem_page[prg_swap() ? 6 : 4]);
  prg1 = MAP95_ROM(context.mem_page[5]);
  if(num_1k_VROM_banks)
  {
    if(chr_swap())
    {
      chr01 = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr23 = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr4  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr5  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
      chr6  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr7  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
    }
    else
    {
      chr01 = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      chr23 = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      chr4  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      chr5  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
      chr6  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      chr7  = MAP95_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
    }
  }
}
/////////////////////////////////////////////////////////////////////

#endif
