
/////////////////////////////////////////////////////////////////////
// Mapper 15
class NES_mapper15 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper15(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper15() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 41
class NES_mapper41 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper41(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper41() {}

  void  Reset();
  void  MemoryWriteSaveRAM(uint32 addr, uint8 data);
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 regs[2];

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 44

class NES_mapper44 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper44(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper44() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);
  void  HSync(uint32 scanline);

protected:
  uint8  regs[8];

  uint32 rom_bank;
  uint32 prg0,prg1;
  uint32 chr01,chr23,chr4,chr5,chr6,chr7;

  uint32 chr_swap() { return regs[0] & 0x80; }
  uint32 prg_swap() { return regs[0] & 0x40; }

  uint8 irq_enabled; // IRQs enabled
  uint8 irq_counter; // IRQ scanline counter, decreasing
  uint8 irq_latch;   // IRQ scanline counter latch

  void MMC3_set_CPU_banks()
  {
    if(prg_swap())
    {
      set_CPU_bank4(16*rom_bank+14);
      set_CPU_bank5(16*rom_bank+prg1);
      set_CPU_bank6(16*rom_bank+prg0);
      set_CPU_bank7(16*rom_bank+15);
    }
    else
    {
      set_CPU_bank4(16*rom_bank+prg0);
      set_CPU_bank5(16*rom_bank+prg1);
      set_CPU_bank6(16*rom_bank+14);
      set_CPU_bank7(16*rom_bank+15);
    }
  }

  void MMC3_set_PPU_banks()
  {
    if(num_1k_VROM_banks)
    {
      if(chr_swap())
      {
        set_PPU_bank0(128*rom_bank+chr4);
        set_PPU_bank1(128*rom_bank+chr5);
        set_PPU_bank2(128*rom_bank+chr6);
        set_PPU_bank3(128*rom_bank+chr7);
        set_PPU_bank4(128*rom_bank+chr01);
        set_PPU_bank5(128*rom_bank+chr01+1);
        set_PPU_bank6(128*rom_bank+chr23);
        set_PPU_bank7(128*rom_bank+chr23+1);
      }
      else
      {
        set_PPU_bank0(128*rom_bank+chr01);
        set_PPU_bank1(128*rom_bank+chr01+1);
        set_PPU_bank2(128*rom_bank+chr23);
        set_PPU_bank3(128*rom_bank+chr23+1);
        set_PPU_bank4(128*rom_bank+chr4);
        set_PPU_bank5(128*rom_bank+chr5);
        set_PPU_bank6(128*rom_bank+chr6);
        set_PPU_bank7(128*rom_bank+chr7);
      }
    }
  }

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 45
class NES_mapper45 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper45(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper45() {}

  void  Reset();
  void  MemoryWriteSaveRAM(uint32 addr, uint8 data);
  void  MemoryWrite(uint32 addr, uint8 data);
  void  HSync(uint32 scanline);

protected:
  uint8  regs[7];

  uint32 p[4],prg0,prg1,prg2,prg3;
  uint32 c[8],chr0,chr1,chr2,chr3,chr4,chr5,chr6,chr7;

  uint8 irq_enabled;
  uint8 irq_counter;
  uint8 irq_latch;

  void MAP45_set_CPU_bank4(uint8 data);
  void MAP45_set_CPU_bank5(uint8 data);
  void MAP45_set_CPU_bank6(uint8 data);
  void MAP45_set_CPU_bank7(uint8 data);
  void MAP45_set_PPU_banks();

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 225
class NES_mapper225 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper225(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper225() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 226
class NES_mapper226 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper226(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper226() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 regs[2];

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 227
class NES_mapper227 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper227(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper227() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 228
class NES_mapper228 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper228(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper228() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 229
class NES_mapper229 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper229(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper229() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 230
class NES_mapper230 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper230(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper230() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 rom_switch;

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 231
class NES_mapper231 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper231(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper231() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 232
class NES_mapper232 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper232(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper232() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 prg_bank0, prg_bank1;

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 233
class NES_mapper233 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper233(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper233() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 234
class NES_mapper234 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper234(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper234() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);
  void  MemoryRead(uint32 addr);

protected:
  void  MemoryWrite2(uint32 addr, uint8 data);
  uint8 regs[3];

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 235
class NES_mapper235 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper235(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper235() {}

  void  Reset();
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 rom_switch;

private:
};
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 255
class NES_mapper255 : public NES_mapper
{
  friend void adopt_MPRD(SnssMapperBlock* block, NES* nes);
  friend int extract_MPRD(SnssMapperBlock* block, NES* nes);

public:
  NES_mapper255(NES* parent) : NES_mapper(parent) {}
  ~NES_mapper255() {}

  void  Reset();
  uint8 MemoryReadLow(uint32);
  void  MemoryWriteLow(uint32 addr, uint8 data);
  void  MemoryWrite(uint32 addr, uint8 data);

protected:
  uint8 regs[4];

private:
};
/////////////////////////////////////////////////////////////////////

